﻿//file: nlgen.cs
//author: Derek Andrews <derek.george.andrews@gmail.com>
//Copyright (C) 2010, All Rights Reserved.
//
//description: 
//This program was written to convert CA65/LD65 listings and
//map file into namelist files consumable by the FCEUXDSP 1.07 emulator.
//I wrote it so that I could view all of my labels and comments (at least
//that which is contained within segments ) while using FCEUXDSP's debugger.
//I thought it could potentially be useful to other NES developers out there
//so I've chosen to release it. If you find any bugs or have any suggestions
//for improvement, please send me an email.
//
//compilation:
//This program ought to compile with Microsoft Visual C#, with .NET 2.0. I'm
//not sure how much of what I'm using in here is implemented in MONO. If you
//ever try to compile this program with MONO on Linux, send me an email and
//let me know. I did start to re-write this utility in C++/Boost, but found
//their regex library is ridiculously slow compared to .NET's. I still have
//the source code lying around if you'd like to see a C++ version for any
//reason.
//
//usage:
// Just pass in the -h switch for help.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;

namespace nlgen
{
    public class Segment
    {
        private string m_namelistKey;

        public string NamelistKey { get { return m_namelistKey; } set { m_namelistKey = value; } }

        private int m_baseAddress;

        public int BaseAddress { get { return m_baseAddress; } set { m_baseAddress = value; } }

        private string m_name;

        public string Name { get { return m_name; } set { m_name = value; } }
    }

    public class NamelistEntry
    {
        private int m_address;

        public int Address { get { return m_address; } set { m_address = value; } }

        private string m_name;

        public string Name { get { return m_name; } set { m_name = value; } }

        private List<string> m_comments;

        public List<string> Comments { get { return m_comments; } set { m_comments = value; } }

        public NamelistEntry()
        {
            m_comments = new List<string>();
        }
    }

    public class NamelistFile
    {
        private Dictionary<int, NamelistEntry> m_namelistEntries;

        public Dictionary<int, NamelistEntry> NamelistEntries { get { return m_namelistEntries; } set { m_namelistEntries = value; } }

        public NamelistFile()
        {
            m_namelistEntries = new Dictionary<int, NamelistEntry>();
        }

        public void addNamelistEntry(int address, NamelistEntry namelistEntry)
        {
            m_namelistEntries.Add(address, namelistEntry);
        }
    }

    public class LstFile
    {
        private Dictionary<string, int> m_segmentOffsets;
        private string m_fileName;

        public string FileName { get { return m_fileName; } set { m_fileName = value; } }

        public LstFile()
        {
            m_segmentOffsets = new Dictionary<string, int>();
        }

        public void addSegmentOffset(string segmentName, int offset)
        {
            m_segmentOffsets[segmentName] = offset;
        }

        public int getSegmentOffset(string segmentName)
        {
            if (m_segmentOffsets.ContainsKey(segmentName))
                return m_segmentOffsets[segmentName];
            else
                return 0;
        }

        public bool containsSegmentOffset(string segmentName)
        {
            return m_segmentOffsets.ContainsKey(segmentName);
        }
    }

    public class nlgen
    {
        private Dictionary<string, NamelistFile> m_namelistFiles;

        private Dictionary<string, Segment> m_segments;

        private Dictionary<string, LstFile> m_lstFiles;

        private string m_mapFile;

        private string m_outputName;

        public nlgen()
        {
            m_namelistFiles = new Dictionary<string, NamelistFile>();
            m_segments = new Dictionary<string, Segment>();
            m_lstFiles = new Dictionary<string, LstFile>();
        }

        bool isValidNamelistKey(string namelistKey)
        {
            Regex regex = new Regex("^ram|[0-9]+$");
            bool success = regex.IsMatch(namelistKey);
            return success;
        }

        bool isValidSegmentName(string segmentName)
        {
            Regex regex = new Regex("^[a-zA-Z0-9_]+$");
            bool success = regex.IsMatch(segmentName);
            return success;
        }

        bool isValidSegmentAddress(string segmentAddress)
        {
            Regex regex = new Regex("^[a-fA-F0-9]{4}$");
            bool success = regex.IsMatch(segmentAddress);
            return success;
        }

        bool isValidFileName(string fileName)
        {
            Regex regex = new Regex(@"^([a-zA-Z0-9._-]+[/\\]{0,1})+$");
            bool success = regex.IsMatch(fileName);
            return success;
        }

        public bool processArgs(string[] args)
        {
            for (int i = 0; i < args.Length; i++)
            {
                if ( args[i] == "-h")
                {
                    Console.WriteLine("nlgen, v. 1.0 by Derek Andrews <derek.george.andrews@gmail.com>");
                    Console.WriteLine();
                    Console.WriteLine("Usage:");
                    Console.WriteLine("-nl NAMELIST_KEY SEGMENT_NAME SEGMENT_ADDRESS ");
                    Console.WriteLine();
                    Console.WriteLine("This switch specifies a segment to fill a namelist file.");
                    Console.WriteLine("The NAMELIST_KEY can be ram or a decimal number.");
                    Console.WriteLine("The SEGMENT_NAME must exist in your project.");
                    Console.WriteLine("The SEGMENT_ADDRESS must be a 4 digit hexadecimal base address.");
                    Console.WriteLine();
                    Console.WriteLine("-lst NAMELIST_FILE_PATH");
                    Console.WriteLine();
                    Console.WriteLine("This switch specifies a namelist file to read.");
                    Console.WriteLine("It expects a relative path to the current directory.");
                    Console.WriteLine("You can specify as many lst files as you wish.");
                    Console.WriteLine();
                    Console.WriteLine("-map MAP_FILE_PATH");
                    Console.WriteLine();
                    Console.WriteLine("This switch specifies the map file to read. This switch is required.");
                    Console.WriteLine();
                    Console.WriteLine("-rom OUTPUT_FILE_PATH");
                    Console.WriteLine();
                    Console.WriteLine("This switch specifies the name of your ROM, e.g. game.nes.");
                    Console.WriteLine("This switch is required in order to generate the correct file names for the output namelist files.");
                    Console.WriteLine();
                    Console.WriteLine("Example usage:");
                    Console.WriteLine();
                    Console.WriteLine("Here is an example from my own game project:");
                    Console.WriteLine();
                    Console.WriteLine("nlgen -rom nomolos.nes -nl ram ZEROPAGE 0000 -nl ram STACK 0100 -nl ram BSS 0200 -nl 0 ROM0 8000 -nl 1 ROM1 8000 -nl 2 ROM2 8000 -nl 7 CODE C000 -map nomolos.map -lst src/nomolos.lst -lst src/nomolosLogic.lst -lst src/rom0.lst -lst src/rom1.lst -lst src/rom2.lst -lst src/chrrom0.lst -lst src/chrrom1.lst -lst src/chrrom2.lst -lst src/chrrom3.lst -lst src/loadLevelState.lst -lst src/playLevelState.lst -lst src/levelOutState.lst -lst src/levelInState.lst -lst src/gameOverState.lst -lst src/titleState.lst -lst src/map.lst -lst src/camera.lst -lst src/sprite.lst -lst src/entity.lst -lst src/controller.lst -lst src/sound.lst -lst src/geotests.lst -lst src/zp.lst -lst src/mapper.lst -lst src/ppu.lst -lst src/gameUIData.lst -lst src/levelDataIndex.lst");
                    return true;
                }
                if (args[i] == "-nl")
                {
                    if (i + 3 < args.Length)
                    {
                        string namelistKey = args[i + 1];
                        string segmentName = args[i + 2];
                        int segmentAddress = 0;
                        if (!isValidSegmentName(segmentName))
                        {
                            Console.WriteLine(string.Format("Segment name {0} is invalid.", segmentName));
                            return false;
                        }
                        if (!isValidSegmentAddress(args[i + 3]))
                        {
                            Console.WriteLine(string.Format("Segment address {0} is invalid. Must be a 4 digit hex number.", args[i + 3]));
                            return false;
                        }
                        Int32.TryParse(args[i + 3], System.Globalization.NumberStyles.HexNumber, null, out segmentAddress);
                        if (isValidNamelistKey(namelistKey))
                        {
                            if (m_namelistFiles.ContainsKey(namelistKey) == false)
                            {
                                NamelistFile namelistFile = new NamelistFile();                                
                                m_namelistFiles.Add(namelistKey, namelistFile);
                            }
                            Segment segment = new Segment();
                            segment.BaseAddress = segmentAddress;
                            segment.NamelistKey = namelistKey;
                            segment.Name = segmentName;
                            m_segments.Add(segmentName, segment);
                        }
                        else
                        {
                            Console.WriteLine("Namelist key invalid.");
                            return false;
                        }
                    }
                    else
                    {
                        Console.WriteLine("-nl switch expects three arguments.");
                        return false;
                    }
                    i += 3;
                }
                if (args[i] == "-lst")
                {
                    if (i + 1 < args.Length)
                    {
                        string lstFileName = args[i + 1];
                        if (isValidFileName(lstFileName))
                        {
                            if (File.Exists(lstFileName))
                            {
                                if (m_lstFiles.ContainsKey(lstFileName) == false)
                                {
                                    string lstFileNameWithoutPath =
                                        Path.GetFileName(lstFileName);
                                    m_lstFiles[lstFileNameWithoutPath] = new LstFile();
                                    m_lstFiles[lstFileNameWithoutPath].FileName = lstFileName;
                                }
                            }
                            else
                            {
                                Console.WriteLine(string.Format(".lst file {0} does not exist.", lstFileName));
                                return false;
                            }
                        }
                        else
                        {
                            Console.WriteLine(string.Format(".lst file {0} is an invalid file name.", lstFileName));
                            return false;
                        }
                    }
                    i += 1;
                }

                if (args[i] == "-map")
                {
                    if (i + 1 < args.Length)
                    {
                        if (isValidFileName(args[i + 1]))
                        {
                            if (File.Exists(args[i + 1]))
                            {
                                m_mapFile = args[i + 1];
                            }
                            else
                            {
                                Console.WriteLine(".map file {0} does not exist.", args[i + 1]);
                                return false;
                            }
                        }
                        else
                        {
                            Console.WriteLine(".map file {0} is an invalid file name.", args[i + 1]);
                            return false;
                        }
                    }
                    else
                    {
                        Console.WriteLine("-map switch expects one argument.");
                        return false;
                    }
                    i += 1;
                }

                if (args[i] == "-rom")
                {
                    if (i + 1 < args.Length)
                    {
                        m_outputName = args[i + 1];
                        if (!isValidFileName(m_outputName))
                        {
                            Console.WriteLine(string.Format("Output name {0} is an invalid file name.", m_outputName));
                            return false;
                        }
                    }
                    else
                    {
                        Console.WriteLine("-rom switch expects one argument.");
                    }
                    i += 1;
                }
            }
            bool fileNotSpecified = false;
            if (m_outputName == null)
            {
                Console.WriteLine("You must specify the name of your rom with -rom ROM_FILE_NAME.");
                fileNotSpecified = true;
            }
            if (m_mapFile == null)
            {
                Console.WriteLine("You must specify the name of your map file with -map MAP_FILE_NAME.");
                fileNotSpecified = true;
            }
            return !fileNotSpecified;
        }

        public bool processMapFile()
        {
            if (m_mapFile == null) return false;
            StreamReader sr = new StreamReader(m_mapFile);
            string currentLstFileName = "";
            while (!sr.EndOfStream)
            {
                string line = sr.ReadLine();

                Regex regex = new Regex("(?<file>[a-zA-Z0-9_.-]+).o:| {4}(?<segment>[a-zA-Z0-9_]*) *Offs = (?<offset>[0-9a-fA-F]{6})   Size = [0-9a-fA-F]{6}");
                int fileGroupNumber = regex.GroupNumberFromName("file");
                int segmentGroupNumber = regex.GroupNumberFromName("segment");
                int offsetGroupNumber = regex.GroupNumberFromName("offset");

                Match match = regex.Match(line);

                Group fileGroup = match.Groups[fileGroupNumber];
                Group segmentGroup = match.Groups[segmentGroupNumber];
                Group offsetGroup = match.Groups[offsetGroupNumber];

                if (match.Success == true)
                {
                    if (fileGroup.Success == true)
                    {
                        currentLstFileName = fileGroup.Value + ".lst";
                    }

                    if (segmentGroup.Success == true && offsetGroup.Success == true)
                    {
                        if (m_lstFiles.ContainsKey(currentLstFileName) == true)
                        {
                            int offsetValue = 0;
                            if (Int32.TryParse(offsetGroup.Value, System.Globalization.NumberStyles.HexNumber, null, out offsetValue) == true)
                            {
                                m_lstFiles[currentLstFileName].addSegmentOffset(segmentGroup.Value, offsetValue);
                            }
                        }
                    }
                }
            }
            sr.Close();
            return true;
        }

        public bool processLstFile(string fileName, LstFile lstFile)
        {
            StreamReader sr = new StreamReader(m_lstFiles[fileName].FileName);
            string currentSegmentName = "";

            Dictionary<int, List<string>> comments = new Dictionary<int, List<string>>();
                        
            while (!sr.EndOfStream)
            {
                string line = sr.ReadLine();

                Regex regex = new Regex("(?<address>[0-9a-fA-F]{6})r [0-9].{15}(( *(?<comment>;.*)| +[a-zA-Z]{3} +.*(?<comment>;.*))|(((?<label>[a-zA-Z0-9_]*:)|(?<label>\\.proc [a-zA-Z0-9_]*))|\\.segment \"(?<segment>.*)\"))");
                int addressGroupNumber = regex.GroupNumberFromName("address");
                int commentGroupNumber = regex.GroupNumberFromName("comment");
                int segmentGroupNumber = regex.GroupNumberFromName("segment");
                int labelGroupNumber = regex.GroupNumberFromName("label");

                Match match = regex.Match(line);
                Group addressGroup = match.Groups[addressGroupNumber];
                Group commentGroup = match.Groups[commentGroupNumber];
                Group segmentGroup = match.Groups[segmentGroupNumber];
                Group labelGroup = match.Groups[labelGroupNumber];

                if (match.Success == true)
                {
                    if (segmentGroup.Success == true)
                        currentSegmentName = segmentGroup.Value;

                    if (m_segments.ContainsKey(currentSegmentName))
                    {
                        if (addressGroup.Success == true)
                        {
                            string addressStr = addressGroup.Value;
                            int address = 0;

                            NamelistFile namelistFile = m_namelistFiles[m_segments[currentSegmentName].NamelistKey];

                            if (Int32.TryParse(addressStr, System.Globalization.NumberStyles.HexNumber, null, out address))
                            {
                                if (m_lstFiles[fileName].containsSegmentOffset(currentSegmentName))
                                {                                
                                    int realAddress = address + m_segments[currentSegmentName].BaseAddress + m_lstFiles[fileName].getSegmentOffset(currentSegmentName);

                                    if ((commentGroup.Success == true || labelGroup.Success == true ) &&
                                        namelistFile.NamelistEntries.ContainsKey(realAddress) == false)
                                    {
                                        NamelistEntry namelistEntry = new NamelistEntry();
                                        namelistEntry.Address = realAddress;
                                        namelistFile.NamelistEntries.Add(realAddress, namelistEntry);
                                    }

                                    if (commentGroup.Success == true)
                                    {
                                        string comment = commentGroup.Value;
                                        namelistFile.NamelistEntries[realAddress].Comments.Add(comment);
                                    }

                                    if (labelGroup.Success == true)
                                    {
                                        string label = labelGroup.Value;
                                        namelistFile.NamelistEntries[realAddress].Name += " " + label;
                                    }
                                }
                                else
                                {
                                    Console.WriteLine("Segment offset for segment {0} not available within .lst file {1}.", currentSegmentName, fileName);
                                }
                            }
                        }
                    }
                }
            }
            sr.Close();

            return true;
        }

        public void GenerateNamelistFiles(string[] args)
        {
            bool processedMapFile = false;
            if (processArgs(args))
            {
                if (processMapFile())
                {
                    Console.WriteLine(string.Format("Successfully processed map file: {0}", m_mapFile));
                    processedMapFile = true;
                }
            }
            if (processedMapFile == true)
            {
                foreach (KeyValuePair<string, LstFile> kvp in m_lstFiles)
                {
                    if (processLstFile(kvp.Key, kvp.Value) == true)
                    {
                        Console.WriteLine(string.Format("Successfully processed lst file: {0}", kvp.Key));
                    }
                }

                foreach (KeyValuePair<string, NamelistFile> kvp in m_namelistFiles)
                {                    
                    string namelistFileName = m_outputName + "." + kvp.Key + ".nl";
                    Console.WriteLine(string.Format("Now writing namelist file: {0}", namelistFileName));
                    StreamWriter sw = new StreamWriter(namelistFileName);
                    NamelistFile namelistFile = kvp.Value;
                    foreach (KeyValuePair<int, NamelistEntry> kvp2 in namelistFile.NamelistEntries)
                    {
                        NamelistEntry namelistEntry = kvp2.Value;

                        if (namelistEntry.Comments.Count == 0)
                        {
                            string namelistEntryLine =
                                string.Format("${0:X4}#{1}#", namelistEntry.Address, namelistEntry.Name);
                            sw.WriteLine(namelistEntryLine);
                        }
                        if (namelistEntry.Comments.Count == 1)
                        {
                            string namelistEntryLine =
                                string.Format("${0:X4}#{1}#{2}", namelistEntry.Address, namelistEntry.Name, namelistEntry.Comments[0]);
                            sw.WriteLine(namelistEntryLine);
                        }
                        if (namelistEntry.Comments.Count > 1)
                        {
                            string namelistEntryLine =
                                string.Format("${0:X4}#{1}#...", namelistEntry.Address, namelistEntry.Name);
                            sw.WriteLine(namelistEntryLine);
                            foreach (string comment in namelistEntry.Comments)
                                sw.WriteLine(string.Format("\\{0}", comment));
                        }
                    }
                    sw.Close();
                }
            }
        }
    }

    class nlgenapp
    {
        static void Main(string[] args)
        {
            nlgen nlgen = new nlgen();
            nlgen.GenerateNamelistFiles(args);
        }
    }
}
